#include <qpushbutton.h>
#include <qcombobox.h>
#include <qradiobutton.h>
#include <qapplication.h>
#include <qspinbox.h>
#include <qsortedlist.h>
#include <qlineedit.h>
#include <qlabel.h>
#include <qcheckbox.h>
#include <qprogressbar.h>
#include <qradiobutton.h>

#include <float.h>


#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/Scene/src/Object.h"
#include "../../../Core/Transformer/src/Transformer.h"
#include "../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../Core/DataStructures/src/SurfelCollection.h"
#include "../../../Core/CoreResourceTools/src/CoreResourceTools.h"
#include "../../../Core/CoreTools/src/CoreTools.h"
#include "../../../DataTypes/src/MyDataTypes.h"
#include "../../../Utilities/src/Matrix.h"
#include "../../../Utilities/src/MyMathTools.h"
#include "../../../Utilities/src/FileTools.h"
#include "../../../Utilities/src/StringTools.h"

#include "Bilateral_DialogImpl.h"


// ******
// global
// ******

// the single Bilateral_DialogImpl instance, only
// accesible by getInstance() and
// destroyInstance()
static Bilateral_DialogImpl *instance = 0;


// *******************
// private c'tor/d'tor
// *******************

Bilateral_DialogImpl::Bilateral_DialogImpl (QWidget *parent, const char *name, bool modal, WFlags flags)
    : Bilateral_Dialog (parent, name, modal, flags) {

	this->initDialog();
}

Bilateral_DialogImpl::~Bilateral_DialogImpl() {
    // no need to delete child widgets, Qt does it all for us
}


// **************
// public methods
// **************

Bilateral_DialogImpl *Bilateral_DialogImpl::getInstance() {

	if (instance == 0) {
		instance = new Bilateral_DialogImpl();
	}
	return instance;

}

void Bilateral_DialogImpl::destroyInstance() {

	if (instance != 0) {
		instance->close();
		delete instance;
	}
	instance = 0;

}

// ***************
// private methods
// ***************

void Bilateral_DialogImpl::initDialog () {
	
	defSpatialWidth = 4.0f;
	SpatialWidthLineEdit->setText (QString::number (defSpatialWidth));
	defInfluenceWidth = 1.0f;
	InfluenceWidthLineEdit->setText (QString::number (defInfluenceWidth));
}



// *************
// private slots
// *************


/* Perform weighted average filtering.
*/
void Bilateral_DialogImpl::handlePointFiltering() {

	unsigned int i, j;
	unsigned int nSurfels;
	Scene *scene;
	SurfelCollection *objectSurfels;
	
	QVector<SurfelInterface> *objectSurfelVector;
	QVector<SurfelInterface> *filteredSurfelVector;
        
        bool jones_predictor = JonesButton->isChecked();

	Vector3D* surfelPositions;
	NeighbourHood *neighborHood;

	Vector3D newPosition;
	float sumWeights, w;

	float radius, area;
	float spatialWidth, influenceWidth;
	bool isNumber;
        float mean_radius;

	scene = Scene::getInstance();

	if (scene->getNofObjects() > 0) {

		QApplication::setOverrideCursor (Qt::waitCursor);
                
		// get the spatial filter variance from the dialog
		spatialWidth = SpatialWidthLineEdit->text().toFloat (&isNumber);
		if (isNumber == false) {
			SpatialWidthLineEdit->setText (QString::number (defSpatialWidth));
			spatialWidth = defSpatialWidth;
		}
		// get the influence filter variance from the dialog
		influenceWidth = InfluenceWidthLineEdit->text().toFloat (&isNumber);
		if (isNumber == false) {
			SpatialWidthLineEdit->setText (QString::number (defInfluenceWidth));
			influenceWidth = defInfluenceWidth;
		}


		// get vector of surfels
		objectSurfels = scene->getActiveObject()->getSurfelCollection();
		objectSurfelVector = objectSurfels->buildSurfelVector();
		nSurfels = objectSurfels->getNofSurfels();

		// get vector of surfel positions
		surfelPositions = new Vector3D[nSurfels];

                // put surfels into positions array, and calculate mean radius.
                mean_radius = 0.0;
		for (i = 0; i < nSurfels; i++) {
			surfelPositions[i] = objectSurfelVector->at(i)->getPosition();
                        mean_radius += objectSurfelVector->at(i)->getRadius();
		}
                mean_radius /= (double) nSurfels;

                spatialWidth *= mean_radius;
                influenceWidth *= mean_radius;

		// allocate vector of filtered surfels
		filteredSurfelVector = new QVector<SurfelInterface>(nSurfels);

		// set up neighborhood
		neighborHood = new NeighbourHood(surfelPositions, nSurfels);
                // radius might be arbitrarily large
		neighborHood->setNofNeighbours (nSurfels);

                // set query distance
                neighborHood->setMaxQuerySqrDistance(2.0 * spatialWidth);

                BilateralProgressBar->setTotalSteps(nSurfels);
                BilateralProgressBar->setProgress(0);
		// filter surfels
		for (i=0; i<nSurfels; i++) {
			neighborHood->setSourcePoint (surfelPositions[i]);
                        unsigned int num_found = neighborHood->getNofFoundNeighbours();

			newPosition = Vector3D(0.f, 0.f, 0.f);
			sumWeights = 0.f;

			for (j = 0; j < num_found; j++) {
                                int neighbor_idx = neighborHood->getNeighbourPositionIndex(j);
                                Vector3D p = surfelPositions[i];
                                Vector3D q = surfelPositions[neighbor_idx];
                                Vector3D prediction, n;

                                // find the area of the current neighbor
                                radius = objectSurfelVector->at(neighbor_idx)->getRadius();
                                // normalization removes factor of 2*PI
                                area = radius * radius; 

                                if (jones_predictor) {
                                  n = objectSurfelVector->at(neighbor_idx)->getNormal();

                                  prediction = p + Vector3D::dotProduct((q - p), n) * n;
                                } else {
                                  n = objectSurfelVector->at(i)->getNormal();

                                  prediction = p + Vector3D::dotProduct((q - p), n) * n;
                                }

				// filter weighted by distance, distance to prediction, and area
				w = area * 
                                  exp(- (p-q).getSquaredLength() / (2.0 * spatialWidth * spatialWidth)) *
                                  exp(- (p - prediction).getSquaredLength() / (2.0 * influenceWidth * influenceWidth));

				newPosition += w * prediction;
				sumWeights += w;
			}

			newPosition /= sumWeights;

			// add a copy of the original surfel to the object surfel collection...
			filteredSurfelVector->insert(i, objectSurfels->addSurfel(objectSurfelVector->at(i), false));
			// ...and adjust its position
			scene->setSurfelPosition(filteredSurfelVector->at(i), newPosition, false);

                        BilateralProgressBar->setProgress(i+1);
		}
                i = 200;
                neighborHood->setSourcePoint (surfelPositions[i]);
                unsigned int num_found = neighborHood->getNofFoundNeighbours();
                for (j = 0; j < num_found; j++) {
                  int neighbor_idx = neighborHood->getNeighbourPositionIndex(j);
                  Vector3D p = surfelPositions[i];
                  Vector3D q = surfelPositions[neighbor_idx];
                  Vector3D prediction, n;
                  
                  // find the area of the current neighbor
                  radius = objectSurfelVector->at(neighbor_idx)->getRadius();
                  // normalization removes factor of 2*PI
                  area = radius * radius; 
                  
                  if (jones_predictor) {
                    n = objectSurfelVector->at(neighbor_idx)->getNormal();
                    
                    prediction = p + Vector3D::dotProduct((q - p), n) * n;
                  } else {
                    n = objectSurfelVector->at(i)->getNormal();
                    
                    prediction = p + Vector3D::dotProduct((q - p), n) * n;
                  }
                  
                  // filter weighted by distance, distance to prediction, and area
                  w = exp(- (p-q).getSquaredLength() / (2.0 * spatialWidth * spatialWidth)) *
                    exp(- (p - prediction).getSquaredLength() / (2.0 * influenceWidth * influenceWidth));
                  

                  scene->setSurfelDiffuseColor(filteredSurfelVector->at(neighbor_idx), QRgb(qRgb((int) (255.0*w),0,0)), false);
                }


		// remove original surfels from object surfel collection
		for (i=0; i<nSurfels; i++) {
			objectSurfels->removeSurfel(objectSurfelVector->at(i), false);
		}

		scene->emitObjectModifiedSignal ();
		scene->emitRenderFrameSignal ();

		QApplication::restoreOverrideCursor ();

		delete objectSurfelVector;
		delete filteredSurfelVector;
		delete[] surfelPositions;
		delete neighborHood;
	}

}


/* Perform normal filtering.
*/
void Bilateral_DialogImpl::handleNormalFiltering() {

    unsigned int i;
    unsigned int nSurfels;
    Scene *scene;

    float spatialWidth, influenceWidth;
    bool isNumber;
    float mean_radius;

    SurfelCollection *objectSurfels;
    QVector<SurfelInterface> *objectSurfelVector;
    QVector<SurfelInterface> *filteredSurfelVector;
    Vector3D* surfelPositions;
    NeighbourHood *neighborHood;

    Vector3D newNormal;

    scene = Scene::getInstance();

	int row, col;

    if (scene->getNofObjects() > 0) {

            QApplication::setOverrideCursor (Qt::waitCursor);

		// get the spatial filter variance from the dialog
		spatialWidth = SpatialWidthLineEdit->text().toFloat (&isNumber);
		if (isNumber == false) {
			SpatialWidthLineEdit->setText (QString::number (defSpatialWidth));
			spatialWidth = defSpatialWidth;
		}
		// get the influence filter variance from the dialog
		influenceWidth = InfluenceWidthLineEdit->text().toFloat (&isNumber);
		if (isNumber == false) {
			SpatialWidthLineEdit->setText (QString::number (defInfluenceWidth));
			influenceWidth = defInfluenceWidth;
		}


		// get vector of surfels
		objectSurfels = scene->getActiveObject()->getSurfelCollection();
		objectSurfelVector = objectSurfels->buildSurfelVector();
		nSurfels = objectSurfels->getNofSurfels();

		// get vector of surfel positions
		surfelPositions = new Vector3D[nSurfels];

                // put surfels into positions array, and calculate mean radius.
                mean_radius = 0.0;
		for (i = 0; i < nSurfels; i++) {
			surfelPositions[i] = objectSurfelVector->at(i)->getPosition();
                        mean_radius += objectSurfelVector->at(i)->getRadius();
		}
                mean_radius /= (double) nSurfels;

                spatialWidth *= mean_radius;
                influenceWidth *= mean_radius;

                // allocate vector of filtered surfels
		filteredSurfelVector = new QVector<SurfelInterface>(nSurfels);

		// set up neighborhood
		neighborHood = new NeighbourHood(surfelPositions, nSurfels);
                // radius might be arbitrarily large
		neighborHood->setNofNeighbours (nSurfels);

                // set query distance
                neighborHood->setMaxQuerySqrDistance(2.0 * spatialWidth);

                BilateralProgressBar->setTotalSteps(nSurfels);
                BilateralProgressBar->setProgress(0);
                // filter surfel normals
                for (i=0; i<nSurfels; i++) {
                  neighborHood->setSourcePoint (surfelPositions[i]);
                  unsigned int j, num_found = neighborHood->getNofFoundNeighbours();

                  double sumWeight = 0.0;
                  Vector3D partialsWeight(0.0, 0.0, 0.0);
                  Vector3D sumPos(0.0, 0.0, 0.0);
                  float jacobianPos[3][3] = {{0.0, 0.0, 0.0},
                                             {0.0, 0.0, 0.0},
                                             {0.0, 0.0, 0.0}};
                  bool jones_predictor = JonesButton->isChecked();
                  double radius, area;


                  if (num_found > 1) {

                    for (j = 0; j < num_found; j++) {
                      int neighbor_idx = neighborHood->getNeighbourPositionIndex(j);
                      Vector3D p = surfelPositions[i];
                      Vector3D q = surfelPositions[neighbor_idx];
                      Vector3D n;
                      
                      if (jones_predictor) {
                        n = objectSurfelVector->at(neighbor_idx)->getNormal();
                      } else {
                        n = objectSurfelVector->at(i)->getNormal();
                      }
                      
                      // prediction
                      Vector3D pi = p + Vector3D::dotProduct((q - p), n) * n;
                      
                      // Jacobian of prediction
                      float jacobianPi[3][3] = {{1-n[0]*n[0],   -n[0]*n[1],  -n[0]*n[2]},
                                                { -n[1]*n[0],  1-n[1]*n[1],  -n[1]*n[2]},
                                                { -n[2]*n[0],   -n[2]*n[1], 1-n[2]*n[2]}};
                      
                      // weight f
                      double f = exp(- (p-q).getSquaredLength() / (2.0 * spatialWidth * spatialWidth));
                      Vector3D partialsF(-2.0 * (p[0] - q[0]) * f / (2.0 * spatialWidth * spatialWidth),
                                         -2.0 * (p[1] - q[1]) * f / (2.0 * spatialWidth * spatialWidth),
                                         -2.0 * (p[2] - q[2]) * f / (2.0 * spatialWidth * spatialWidth));

                      // weight g
                      double g = exp(- (p-pi).getSquaredLength() / (2.0 * influenceWidth * influenceWidth));
                      Vector3D partialsG(-2.0 * Vector3D::dotProduct(p - q, n) * n[0] * g / (2.0 * influenceWidth * influenceWidth),
                                         -2.0 * Vector3D::dotProduct(p - q, n) * n[1] * g / (2.0 * influenceWidth * influenceWidth),
                                         -2.0 * Vector3D::dotProduct(p - q, n) * n[2] * g / (2.0 * influenceWidth * influenceWidth));

                      // find the area of the current neighbor
                      radius = objectSurfelVector->at(neighbor_idx)->getRadius();
                      // normalization removes factor of 2*PI
                      area = radius * radius; 
                  

                      // running totals
                      sumWeight += f * g * area;
                      partialsWeight += area * (partialsF * g + f * partialsG);
                      sumPos += f * g * area * pi;
                      for (row = 0; row < 3; row++) {
                        for (col = 0; col < 3; col++) {
                          jacobianPos[row][col] += area * (partialsF[col] * g * pi[row] + 
                                                           f * partialsG[col] * pi[row] +
                                                           f * g * jacobianPi[row][col]);
                        }
                      }
                    }

                    float jacobianPred[3][3];
                    for (row = 0; row < 3; row++) {
                      for (col = 0; col < 3; col++) {
                        // since we don't care about arbitrary scales in this matrix, drop one factor of (1/sumWeight)
                        jacobianPred[row][col] = jacobianPos[row][col] - (sumPos[row] * partialsWeight[col] / sumWeight);
                      }
                    }

                    float adjointPred[3][3], tradjointPred[3][3];
                    MtrAdjoint3x3f(&(jacobianPred[0][0]), &(adjointPred[0][0]));
                    MtrTranspose3x3f(&(adjointPred[0][0]), &(tradjointPred[0][0]));
                    Vector3D oldNormal =  objectSurfelVector->at(i)->getNormal();
                    for (row = 0; row < 3; row++) {
                      newNormal[row] = (tradjointPred[row][0] * oldNormal[0] +
                                        tradjointPred[row][1] * oldNormal[1] +
                                        tradjointPred[row][2] * oldNormal[2]);
                    }
                    if (newNormal.getSquaredLength() > 0.0) {
                      newNormal.normalize();
                      if (Vector3D::dotProduct(newNormal, oldNormal) < 0.0) {
                        newNormal *= -1.0;
                        printf("flipped!\n");
                        for (int row = 0; row < 3; row++) {
                          printf("  %f %f %f\n", tradjointPred[row][0], tradjointPred[row][1], tradjointPred[row][2]);
                        }
                      }
                    } else {
                      printf("degenerate!\n");
                      newNormal = oldNormal;
                    }

                    // add a copy of the original surfel to the object surfel collection...
                    filteredSurfelVector->insert(i, objectSurfels->addSurfel(objectSurfelVector->at(i), false));
                    // ...and adjust its position
                    scene->setSurfelNormal(filteredSurfelVector->at(i), newNormal, false);
                  }
                  BilateralProgressBar->setProgress(i+1);
                }

                // remove original surfels from object surfel collection
		for (i=0; i<nSurfels; i++) {
                  objectSurfels->removeSurfel(objectSurfelVector->at(i), false);
		}

                scene->emitObjectModifiedSignal ();
                scene->emitRenderFrameSignal ();
                
                QApplication::restoreOverrideCursor ();

                delete objectSurfelVector;
		delete filteredSurfelVector;
                delete[] surfelPositions;
                delete neighborHood;
    }
}

